#include "huhfa.h"

huhfa::huhfa(string source){
    //Auxiliary Variables
    map<string, int> index;     //translates variables into their corresponding matrix-columns
    vector<int> vertex;
    int intbuff;
    char charbuff;
    string stringbuff;
    string variable;
    char sign;
    char lookup;
    int value;

    //Filenames
    string PortaIn=source;
    PortaIn.append(".poi");
    string PortaOut=source;
    PortaOut.append(".poi.ieq");
    string Symmetry=source;
    Symmetry.append(".huh");

    /*****************************************************************************************************/
    /** PORTA IN **/
    /*****************************************************************************************************/
    ifstream file1(PortaIn.c_str(), ios::in);
    if(!file1){
        cerr<<"File error: File "<<PortaIn<<" not found."<<endl;
        input_fail=true; return;
    }
    //ignore "DIM =" or "DIM= " or "DIM="
    lookup='#';
     while(lookup != '='){
         file1>>lookup;
     }
    //read dimension
    file1>>n;
    if(file1.fail()){
        cerr<<"Parsing error: Dimension in "<<PortaIn<<" could not be read."<<endl;
        input_fail=true; return;
    }

    //ignore "INDEX"
    file1>>stringbuff;
    //store variable names in an array
    for(int i=0; i<n; i++){
        file1>>stringbuff;
        variables.push_back(stringbuff);
    }
    if(file1.fail()){
        cerr<<"Parsing error: Variables in "<<PortaIn<<" could not be read."<<endl;
        input_fail=true; return;
    }
    //ignore "CONV_SECTION"
    file1>>stringbuff;
    //parse the document for "END"
    for(file1>>stringbuff; (stringbuff != "END") && (!file1.fail()); file1>>stringbuff){
        //every line becomes a vertex
        vertex.clear();
        stringstream ss(stringbuff);
        ss>>intbuff;
        vertex.push_back(intbuff);
        for(int i=1; i<n; i++){
            file1>>intbuff;
            vertex.push_back(intbuff);
        }
        vertices.push_back(vertex);
    }
    if(file1.fail()){
        cerr<<"Parsing error: Vertices in "<<PortaIn<<" do not match data format."<<endl;
        input_fail=true; return;
    }

    /*****************************************************************************************************/
    /** PORTA OUT **/
    /*****************************************************************************************************/
    ifstream file2(PortaOut.c_str(), ios::in);
    if(!file2){
        cerr<<"File error: File "<<PortaOut<<" not found."<<endl;
        input_fail=true; return;
    }

    /** PARSE DIMENSION **/

    //ignore "DIM" and "="
    lookup='#';
     while(lookup != '='){
         file2>>lookup;
     }
    //read dimension
    file2>>intbuff;
    if(file2.fail()){
        cerr<<"Parsing error: Dimension in "<<PortaOut<<" could not be read."<<endl;
        input_fail=true; return;
    }
    if(intbuff != n){
        cerr<<"Input error: Mismatch in dimension between "<<PortaIn<< " and "<<PortaOut<<"."<<endl;
        input_fail=true; return;
    }

    /** PARSE VARIABLES **/

    //ignore "INDEX"
    file2>>stringbuff;
    //store table of variable names and define their corresponding matrix column as index
    for(int i=0; i<n; i++){
        file2>>stringbuff;
        if(stringbuff != variables[i]){
            cerr<<"Input Error: Mismatch in variables between "<<PortaIn<< " and "<<PortaOut<<"."<<endl;
            input_fail=true; return;
        }
        index[stringbuff]=i;
    }
    if(file2.fail()){
        cerr<<"Parsing error: Variables in "<<PortaOut<<" could not be read."<<endl;
        input_fail=true; return;
    }


    /** PARSE INEQUALITIES **/
    //Parse 'Inequalities'
    file2>>stringbuff;
    //start parsing
    vector<int> inequality(n+1,0);
    //parse the document for "VALID" or "END"
    for(file2>>stringbuff; (stringbuff != "VALID") && (stringbuff != "END") && (!file2.fail()) ; file2>>stringbuff){
        //ignore the number of the current inequality
        if(stringbuff[0]=='('){
            while(stringbuff[stringbuff.length()-1] != ')'){
                file2>>stringbuff;
            }
        }
        //parse a single inequality
        for( ; (stringbuff != "<=") && (!file2.fail()); file2>>stringbuff){
            //appending "+" as a delimiter for the last summand
            stringbuff.append("+");
            //tokenize the string of a sum without whitespace into summands
            size_t start = stringbuff.find_first_of("+-");
            for(size_t end = stringbuff.find_first_of("+-", stringbuff.find_first_not_of("+-",start)); end != string::npos; end= stringbuff.find_first_of("+-", stringbuff.find_first_not_of("+-",start))){
                //convert the token into the corresponding matrix element
                stringstream token(stringbuff.substr(start, (end-start)));
                value=1;
                token>>sign;
                if(token.fail()){
                    cerr<<"Parsing error: Inequalities in "<<PortaOut<<" do not match data format."<<endl;
                    input_fail=true; return;
                }
                lookup=token.peek();
                if( (lookup>='0') && (lookup<='9') ){
                    token>>value;
                    if(token.fail()){
                        cerr<<"Parsing error: Inequalities in "<<PortaOut<<" do not match data format."<<endl;
                        input_fail=true; return;
                    }
                }
                if(sign=='-'){
                    value*=-1;
                }
                token>>variable;
                if(token.fail()){
                    cerr<<"Parsing error: Inequalities in "<<PortaOut<<" do not match data format."<<endl;
                    input_fail=true; return;
                }
                inequality[index[variable]]=value;
                start = end;
            }
        }
        //rhs of the equation
        file2>>inequality[n];
        sort_ineq.insert(inequality);
        //reset of the inequality vector
        for(int i=0; i<=n; i++){
            inequality[i]=0;
        }
    }
    if(file2.fail()){
        cerr<<"Parsing error: Inequalities in "<<PortaOut<<" do not match data format.."<<endl;
        input_fail=true; return;
    }

    /*****************************************************************************************************/
    /** SYMMETRY **/
    /*****************************************************************************************************/
    ifstream file3(Symmetry.c_str(), ios::in);
    if(!file3){
        cerr<<"File Error: File "<<Symmetry<<" not found."<<endl;
        input_fail=true; return;
    }
    //parse the affine maps until eof and count their number in m
    for(p=0; !file3.fail(); p++){
        //make a new map
        vector<vector<pair<int, int> > > affinemap;
        //tokenize the sum of every entry and convert them to maps
        for(file3>>stringbuff; (stringbuff != ";") && (!file3.fail()); file3>>stringbuff){
            vector<pair<int, int> > affineimage;
            //appending "+" as a delimiter for the last summand
            stringbuff.append("+");
            //tokenize the string of a sum without whitespace into summands
            size_t start = stringbuff.find_first_of("+-");
            for(size_t end = stringbuff.find_first_of("+-", stringbuff.find_first_not_of("+-",start)); end != string::npos; end= stringbuff.find_first_of("+-", stringbuff.find_first_not_of("+-",start))){
                //convert the token into the corresponding matrix element
                stringstream token(stringbuff.substr(start, (end-start)));
                value=1;
                token>>sign;
                if(token.fail()){
                    cerr<<"Parsing error: Symmetrys in "<<Symmetry<<" do not match data format."<<endl;
                    input_fail=true; return;
                }
                lookup=token.peek();
                if( (lookup>='0') && (lookup<='9') ){
                    token>>value;
                    if(token.fail()){
                        cerr<<"Parsing error: Symmetrys in "<<Symmetry<<" do not match data format."<<endl;
                        input_fail=true; return;
                    }
                }
                if(sign=='-'){
                    value*=-1;
                }
                token>>variable;
                if(token.fail()){
                     affineimage.push_back(make_pair(-1,value));
                     break;
                }
                affineimage.push_back(make_pair(index[variable],value));
                start = end;
            }
            affinemap.push_back(affineimage);
        }
        if(affinemap.size()==n){
            affinemaps.push_back(affinemap);
        }
        else if(affinemap.size()>0){
            cerr<<"Warning: Some symmetry in "<<Symmetry<<" have not enough entries and have been skipped"<<endl;
        }
    }
    //m has to be decremented as it was incremented although there was no symmetry added at the end of the loop
    p--;
    v=vertices.size();
    f=sort_ineq.size();
    input_fail=false; return;
}


